using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using Pathfinding;
using Pathfinding.Serialization.JsonFx;

namespace Pathfinding {
	
	/** A class for holding a user placed connection */
	public class UserConnection {
		
		public Vector3 p1;
		public Vector3 p2;
		public ConnectionType type;
		
		//Connection
		[JsonName("doOverCost")]
		public bool doOverrideCost = false;
		[JsonName("overCost")]
		public int overrideCost = 0;
		
		public bool oneWay = false;
		public bool enable = true;
		public float width = 0;
		
		//Modify Node
		[JsonName("doOverWalkable")]
		public bool doOverrideWalkability = true;
		
		[JsonName("doOverCost")]
		public bool doOverridePenalty = false;
		[JsonName("overPenalty")]
		public uint overridePenalty = 0;
		
	}
	
	[System.Serializable]
	/** Stores editor colors */
	public class AstarColor {
		
		public Color _NodeConnection;
		public Color _UnwalkableNode;
		public Color _BoundsHandles;
	
		public Color _ConnectionLowLerp;
		public Color _ConnectionHighLerp;
		
		public Color _MeshEdgeColor;
		public Color _MeshColor;
		
		/** Holds user set area colors.
		 * Use GetAreaColor to get an area color */
		public Color[] _AreaColors;
		
		public static Color NodeConnection = new Color (1,1,1,0.9F);
		public static Color UnwalkableNode = new Color (1,0,0,0.5F);
		public static Color BoundsHandles = new Color (0.29F,0.454F,0.741F,0.9F);
		
		public static Color ConnectionLowLerp = new Color (0,1,0,0.5F);
		public static Color ConnectionHighLerp = new Color (1,0,0,0.5F);
		
		public static Color MeshEdgeColor = new Color (0,0,0,0.5F);
		public static Color MeshColor = new Color (0,0,0,0.5F);
		
		/** Holds user set area colors.
		 * Use GetAreaColor to get an area color */
		private static Color[] AreaColors;
		
		/** Returns an color for an area, uses both user set ones and calculated.
		 * If the user has set a color for the area, it is used, but otherwise the color is calculated using Mathfx::IntToColor
		 * \see #AreaColors */
		public static Color GetAreaColor (int area) {
			if (AreaColors == null || area >= AreaColors.Length) {
				return Mathfx.IntToColor (area,1F);
			}
			return AreaColors[area];
		}
		
		/** Pushes all local variables out to static ones */
		public void OnEnable () {
			
			NodeConnection = _NodeConnection;
			UnwalkableNode = _UnwalkableNode;
			BoundsHandles = _BoundsHandles;
			
			ConnectionLowLerp = _ConnectionLowLerp;
			ConnectionHighLerp = _ConnectionHighLerp;
			
			MeshEdgeColor = _MeshEdgeColor;
			MeshColor = _MeshColor;
			
			AreaColors = _AreaColors;
		}
		
		public AstarColor () {
			
			_NodeConnection = new Color (1,1,1,0.9F);
			_UnwalkableNode = new Color (1,0,0,0.5F);
			_BoundsHandles = new Color (0.29F,0.454F,0.741F,0.9F);
			
			_ConnectionLowLerp = new Color (0,1,0,0.5F);
			_ConnectionHighLerp = new Color (1,0,0,0.5F);
			
			_MeshEdgeColor = new Color (0,0,0,0.5F);
			_MeshColor = new Color (0.125F, 0.686F, 0, 0.19F);
		}
		
		//new Color (0.909F,0.937F,0.243F,0.6F);
	}
	
	
	/** Returned by graph ray- or linecasts containing info about the hit. This will only be set up if something was hit. \todo Why isn't this a struct? */
	public class GraphHitInfo {
		public Vector3 origin;
		public Vector3 point;
		public Node node;
		public Vector3 tangentOrigin;
		public Vector3 tangent;
		public bool success;
		
		public float distance {
			get {
				return (point-origin).magnitude;
			}
		}
		
		public GraphHitInfo () {
			success = false;
			tangentOrigin  = Vector3.zero;
			origin = Vector3.zero;
			point = Vector3.zero;
			node = null;
			tangent = Vector3.zero;
		}
		
		public GraphHitInfo (Vector3 point) {
			success = false;
			tangentOrigin  = Vector3.zero;
			origin = Vector3.zero;
			this.point = point;
			node = null;
			tangent = Vector3.zero;
			//this.distance = distance;
		}
	}
	
	/** Nearest node constraint. Constrains which nodes will be returned by the GetNearest function */
	public class NNConstraint {
		
		/** Graphs treated as valid to search on.
		 * This is a bitmask meaning that bit 0 specifies whether or not the first graph in the graphs list should be able to be included in the search,
		 * bit 1 specifies whether or not the second graph should be included and so on.
		 * \code
		 * //Enables the first and third graphs to be included, but not the rest
		 * myNNConstraint.graphMask = (1 << 0) | (1 << 2);
		 * \endcode
		 * \note This does only affect which nodes are returned from a GetNearest call, if an invalid graph is linked to from a valid graph, it might be searched anyway.
		 * 
		 * \see AstarPath::GetNearest */
		public int graphMask = -1;
		
		public bool constrainArea = false; /**< Only treat nodes in the area #area as suitable. Does not affect anything if #area is less than 0 (zero) */ 
		public int area = -1; /**< Area ID to constrain to. Will not affect anything if less than 0 (zero) or if #constrainArea is false */
		
		public bool constrainWalkability = true; /**< Only treat nodes with the walkable flag set to the same as #walkable as suitable */
		public bool walkable = true; /**< What must the walkable flag on a node be for it to be suitable. Does not affect anything if #constrainWalkability if false */
		
		public bool constrainTags = true; /**< Sets if tags should be constrained */
		public int tags = -1; /**< Nodes which have any of these tags set are suitable. This is a bitmask, i.e bit 0 indicates that tag 0 is good, bit 3 indicates tag 3 is good etc. */
		
		/** Constrain distance to node.
		 * Uses distance from AstarPath.maxNearestNodeDistance.
		 * If this is false, it will completely ignore the distance limit.
		 * \note This value is not used in this class, it is used by the AstarPath.GetNearest function.
		 */
		public bool constrainDistance = true;
		
		/** Returns whether or not the graph conforms to this NNConstraint's rules.
		  */
		public virtual bool SuitableGraph (int graphIndex, NavGraph graph) {
			return ((graphMask >> graphIndex) & 1) != 0;
		}
		
		/** Returns whether or not the node conforms to this NNConstraint's rules */
		public virtual bool Suitable (Node node) {
			if (constrainWalkability && node.walkable != walkable) return false;
			
			if (constrainArea && area >= 0 && node.area != area) return false;
			
			if (constrainTags && (tags >> node.tags & 0x1) == 0) return false;
			
			return true;
		}
		
		/** The default NNConstraint */
		public static NNConstraint Default {
			get {
				return new NNConstraint ();
			}
		}
		
		/** Returns a constraint which will not filter the results */
		public static NNConstraint None {
			get {
				NNConstraint n = new NNConstraint ();
				n.constrainWalkability = false;
				n.constrainArea = false;
				n.constrainTags = false;
				n.constrainDistance = false;
				n.graphMask = -1;
				return n;
			}
		}
		
		/** Default constructor. Equals to the property #Default */
		public NNConstraint () {
		}
	}
	
	/** A special NNConstraint which can use different logic for the start node and end node in a path.
	 * A PathNNConstraint can be assigned to the Path::nnConstraint field, the path will first search for the start node, then it will call #SetStart and proceed with searching for the end node (nodes in the case of a MultiTargetPath).\n
	 * The #Default PathNNConstraint will constrain the end point to lie inside the same area as the start point.
	 */
	public class PathNNConstraint : NNConstraint {
		
		public static new PathNNConstraint Default {
			get {
				PathNNConstraint n = new PathNNConstraint ();
				n.constrainArea = true;
				return n;
			}
		}
		
		/** Called after the start node has been found. This is used to get different search logic for the start and end nodes in a path */
		public virtual void SetStart (Node node) {
			if (node != null) {
				area = node.area;
			} else {
				constrainArea = false;
			}
		}
	}
	
	public class NNInfo {
		public Node node;
		
		/** Optional to be filled in. if the search will be able to find the constrained node without any extra effort it can fill it in. */
		public Node constrainedNode;
		
		public NearestNodePriority priority = NearestNodePriority.Normal;
		
		public Vector3 clampedPosition;
		/** Clamped position for the optional constrainedNode */
		public Vector3 constClampedPosition;
		
		public NNInfo () {}
		
		public NNInfo (Node node) {
			priority = NearestNodePriority.Normal;
			this.node = node;
			constrainedNode = null;
			constClampedPosition = Vector3.zero;
			
			if (node != null) {
				clampedPosition = (Vector3)node.position;
			} else {
				clampedPosition = Vector3.zero;
			}
		}
		
		/** Sets the constrained node */
		public void SetConstrained (Node constrainedNode, Vector3 clampedPosition) {
			this.constrainedNode = constrainedNode;
			constClampedPosition = clampedPosition;
		}
		
		/** Updates #clampedPosition and #constClampedPosition from node positions */
		public void UpdateInfo () {
			if (node != null) {
				clampedPosition = (Vector3)node.position;
			} else {
				clampedPosition = Vector3.zero;
			}
			
			if (constrainedNode != null) {
				constClampedPosition = (Vector3)constrainedNode.position;
			} else {
				constClampedPosition = Vector3.zero;
			}
		}
		
		public NNInfo (Node node, NearestNodePriority priority) {
			this.node = node;
			this.priority = priority;
			clampedPosition = (Vector3)node.position;
			constrainedNode = null;
			constClampedPosition = Vector3.zero;
		}
		
		public static implicit operator Vector3 (NNInfo ob) {
			return ob.clampedPosition;
		}
		
		public static implicit operator Node (NNInfo ob) {
			return ob.node;
		}
		
		public static implicit operator NNInfo (Node ob) {
			return new NNInfo (ob);
		}
	}
	
	public struct Progress {
		public float progress;
		public string description;
		
		public Progress (float p, string d) {
			progress = p;
			description = d;
		}
	}
	
	public interface IUpdatableGraph {
		
		/** Updates an area using the specified GraphUpdateObject.
		 * 
		 * Notes to implementators.
		 * This function should (in order):
		 * -# Call o.WillUpdateNode on the GUO for every node it will update, it is important that this is called BEFORE any changes are made to the nodes.
		 * -# Update walkabilty using special settings such as the usePhysics flag used with the GridGraph.
		 * -# Call Apply on the GUO for every node which should be updated with the GUO.
		 * -# Update eventual connectivity info if appropriate (GridGraphs updates connectivity, but most other graphs don't since then the connectivity cannot be recovered later).
		 */
		void UpdateArea (GraphUpdateObject o);
	}
	
	//Enumerators
	
	/** Used to weight nodes returned by the GetNearest function from different graphs */
	public enum NearestNodePriority {
		ReallyLow = 20, 	
		Low = 8,		/**< Used when only a simple range check or a similar algorithm was used */
		Normal = 5, 	/**< Default */
		High = 1,		/**< Used when the node is a very good candidate to being the closest node (like when the position is inside a triangle on a navmesh), it is really hard to override this */
		ReallyHigh = 0 	/**< No other NavGraph can override this except if another graph returned ReallyHigh before this one */
	}
	
	[System.Serializable]
	/** Holds a tagmask.
	 * This is used to store which tags to change and what to set them to in a Pathfinding::GraphUpdateObject.
	 * All variables are bitmasks.\n
	 * I wanted to make it a struct, but due to technical limitations when working with Unity's GenericMenu, I couldn't.
	 * So be wary of this when passing it as it will be passed by reference, not by value as e.g LayerMask.
	 */
	public class TagMask {
		public int tagsChange;
		public int tagsSet;
		
		public TagMask () {}
		
		public TagMask (int change, int set) {
			tagsChange = change;
			tagsSet = set;
		}
		
		public void SetValues (System.Object boxedTagMask) {
			TagMask o = (TagMask)boxedTagMask;
			tagsChange = o.tagsChange;
			tagsSet = o.tagsSet;
			//Debug.Log ("Set tags to "+tagsChange +" "+tagsSet+" "+someVar);
		}
		
		public override string ToString () {
			return ""+System.Convert.ToString (tagsChange,2)+"\n"+System.Convert.ToString (tagsSet,2);
		}
	}
	
	/** Represents a collection of settings used to update nodes in a specific area of a graph.
	 * \see AstarPath::UpdateGraphs
	 */
	public class GraphUpdateObject {
		
		/** The bounds to update nodes within */
		public Bounds bounds;
		
		/** Performance boost.
		 * This controlls if a flood fill will be carried out after this GUO has been applied.\n
		 * If you are sure that a GUO will not modify walkability or connections. You can set this to false.
		 * For example when only updating penalty values it can save processing power when setting this to false. Especially on large graphs.
		 * \note If you set this to false, even though it does change e.g walkability, it can lead to paths returning that they failed even though there is a path,
		 * or the try to search the whole graph for a path even though there is none, and will in the processes use wast amounts of processing power.
		 *
		 * If using the basic GraphUpdateObject (not a derived class), a quick way to check if it is going to need a flood fill is to check if #modifyWalkability is true or #updatePhysics is true.
		 *
		 */
		public bool requiresFloodFill = true;
		
		/** Use physics checks to update nodes.
		 * When updating a grid graph and this is true, the nodes' position and walkability will be updated using physics checks
		 * with settings from "Collision Testing" and "Height Testing".
		 * Also when updating a PointGraph, setting this to true will make it re-evaluate all connections in the graph which passes through the #bounds.
		 * This has no effect when updating GridGraphs if #modifyWalkability is turned on */
		public bool updatePhysics = true;
		
		/** NNConstraint to use.
		 * The Pathfinding::NNConstraint::SuitableGraph function will be called on the NNConstraint to enable filtering of which graphs to update.\n
		 * \note As the Pathfinding::NNConstraint::SuitableGraph function is A* Pathfinding Project Pro only, this variable doesn't really affect anything in the free version.
		 * 
		 * \astarpro */
		public NNConstraint nnConstraint = NNConstraint.None;
		
		/** Penalty to add to the nodes */
		public int addPenalty = 0;
		
		public bool modifyWalkability = false; /**< If true, all nodes \a walkable variables will be set to #setWalkability */
		public bool setWalkability = false; /**< If #modifyWalkability is true, the nodes' \a walkable variable will be set to this */
		
		public bool modifyTag = false;
		public int setTag = 0;
		
		public bool trackChangedNodes = false;
		
		private List<Node> changedNodes;
		private List<ulong> backupData;
		private List<Int3> backupPositionData;
		
		public GraphUpdateShape shape = null;
		
		/** Should be called on every node which is updated with this GUO before it is updated */
		public virtual void WillUpdateNode (Node node) {
			if (trackChangedNodes) {
				if (changedNodes == null) { changedNodes = new List<Node>(); backupData = new List<ulong>(); backupPositionData = new List<Int3>(); }
				changedNodes.Add (node);
				backupPositionData.Add (node.position);
				backupData.Add ((ulong)node.penalty<<32 | (ulong)node.flags);
			}
		}
		
		/** Reverts penalties and flags (which includes walkability) on every node which was updated using this GUO.
		 * Data for reversion is only saved if #trackChangedNodes is true */
		public virtual void RevertFromBackup () {
			if (trackChangedNodes) {
				if (changedNodes == null) return;
				for (int i=0;i<changedNodes.Count;i++) {
					changedNodes[i].penalty = (uint)(backupData[i]>>32);
					changedNodes[i].flags = (int)(backupData[i] & 0xFFFFFFFF);
					changedNodes[i].position = backupPositionData[i];
				}
			} else {
				throw new System.InvalidOperationException ("Changed nodes have not been tracked, cannot revert from backup");
			}
		}
		
		/** Updates the specified node using this GUO's settings */
		public virtual void Apply (Node node) {
			if (shape == null || shape.Contains	(node)) {
				
				//Update penalty and walkability
				node.penalty = (uint)(node.penalty+addPenalty);
				if (modifyWalkability) {
					node.walkable = setWalkability;
				}
				
				//Update tags
				if (modifyTag) node.tags = setTag;
			}
		}
		
		public GraphUpdateObject () {
		}
		
		/** Creates a new GUO with the specified bounds */
		public GraphUpdateObject (Bounds b) {
			bounds = b;
		}
	}
	
	public interface IRaycastableGraph {
		bool Linecast (Vector3 start, Vector3 end);
		bool Linecast (Vector3 start, Vector3 end, Node hint);
		bool Linecast (Vector3 start, Vector3 end, Node hint, out GraphHitInfo hit);
	}
	
	/** Holds info about one pathfinding thread.
	 * Mainly used to send information about how the thread should execute when starting it
	  */
	public struct PathThreadInfo {
		public int threadIndex;
		public AstarPath astar;
		public NodeRunData runData;
		
		private System.Object _lock;
		public System.Object Lock { get {return _lock; }}
		
		public PathThreadInfo (int index, AstarPath astar, NodeRunData runData) {
			this.threadIndex = index;
			this.astar = astar;
			this.runData = runData;
			_lock = new object();
		}
	}
}

//Delegates
	
//public delegate (Path p

/* Delegate with on Path object as parameter.
 * This is used for callbacks when a path has finished calculation.\n
 * Example function:
 * \code
public void Start () {
	//Assumes a Seeker component is attached to the GameObject
	Seeker seeker = GetComponent<Seeker>();
	
	//seeker.pathCallback is a OnPathDelegate, we add the function OnPathComplete to it so it will be called whenever a path has finished calculating on that seeker
	seeker.pathCallback += OnPathComplete;
}

public void OnPathComplete (Path p) {
	Debug.Log ("This is called when a path is completed on the seeker attached to this GameObject");
}\endcode
  */
public delegate void OnPathDelegate (Path p);

public delegate Vector3[] GetNextTargetDelegate (Path p, Vector3 currentPosition);

//public delegate void OnPathSucess (Path p);

//public delegate void OnPathError (Path p);

//public delegate void OnPathPreSearch (Path p);

//public delegate void OnPathPostSearch (Path p);

public delegate void OnGraphDelegate (NavGraph graph);

//public delegate void OnGraphPostScan (NavGraph graph);

public delegate void OnScanDelegate (AstarPath script);

public delegate void OnVoidDelegate ();

//public delegate void OnPostScan ();

/** How path results are logged by the system */
public enum PathLog {
	None,		/**< Does not log anything */
	Normal,		/**< Logs basic info about the paths */
	Heavy,		/**< Includes additional info */
	InGame,		/**< Same as heavy, but displays the info in-game using GUI */
	OnlyErrors	/**< Same as normal, but logs only paths which returned an error */
}

/** Heuristic to use. Heuristic is the estimated cost from the current node to the target */
public enum Heuristic {
	Manhattan,
	DiagonalManhattan,
	Euclidean,
	None
}

/** What data to draw the graph debugging with */
public enum GraphDebugMode {
	Areas,
	G,
	H,
	F,
	Penalty,
	Connections,
	Tags
}

/** Type of connection for a user placed link */
public enum ConnectionType {
	Connection,
	ModifyNode
}

public enum ThreadCount {
	Automatic = -1,
	None = 0,
	One = 1,
	Two,
	Three,
	Four,
	Five,
	Six,
	Seven,
	Eight
}