using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using Pathfinding;

public class UnitFactory : MonoBehaviourHeritor
{      
    [System.Serializable]
    public class CreatingUnit_Data
    {
        public GameObject Prefab 
        {
            get { return _prefab; } 
        }

        public int Price
        {
            get { return _price; }
        }

        public float CreatingDuration
        {
            get { return _creatingDuration; }
        }

        public GameObject _prefab;
        public int _price;
        public float _creatingDuration;

        public CreatingUnit_Data(GameObject prefab, int price, float creatingDuration)
        {
            if (!prefab.IsUnit()) //    
                Debug.LogError("Prefab has not "+Tags.Unit+" tag");

            _prefab = prefab;
            _price = price;
            _creatingDuration = creatingDuration;
        }
    }

    #region Inspector Variables

    [SerializeField]
    Vector3 _spawnPoint; //     

    [SerializeField]
    Vector3 _spawnRotation; //     
    
    #endregion


    public Vector3 AssemblyPoint { get; set; }
    public Player OwnerPlayer { get; private set; }

    //    SerializeField    
    [SerializeField]
    List<CreatingUnit_Data> _taskDataList = new List<CreatingUnit_Data>();

    [SerializeField] 
    float _timeBeforeSpawn=0;

    [SerializeField] 
    CreatingUnit_Data _currentTaskData;

    bool _hasCreatingUnit;

    public void AddUnitToFactory(CreatingUnit_Data data)
    {
        if (OwnerPlayer == null)
        {
            OwnerPlayer = GameObjectManager.GetOwnerPlayer(gameObject);
            _currentTaskData = null;
        }

        if (data != null)
        {
            if (OwnerPlayer.Money - data.Price >= 0) // ,      
            {
                OwnerPlayer.Money -= data.Price;
                _taskDataList.Add(data);
                enabled = true;
                _hasCreatingUnit = true;
                if (_timeBeforeSpawn <=0)
                    DequeueUnitData();
            }
        }
    }

    public void CancelUnitCreating(int unitNumber)
    { 
    }

    void Start()
    {
        /*Vector3 pos = transform.TransformPoint(_spawnPoint);
        AssemblyPoint = pos;

        Quaternion rot = new Quaternion();
        rot.eulerAngles = _spawnRotation;
        Vector3 dir;
        rot = transform.rotation * rot;

        //forward
        dir = (rot * Vector3.forward).normalized;
        AssemblyPoint += dir;*/
    }

    void Update()
    {
        if (_hasCreatingUnit)
        {
            _timeBeforeSpawn -= Time.deltaTime;
            if (_timeBeforeSpawn <= 0)
            {
                Quaternion unitRotation = new Quaternion();
                unitRotation.eulerAngles = _spawnRotation;
                unitRotation = transform.rotation * unitRotation;

                UnitAI newUnit = GameObjectManager.CreateObject(_currentTaskData.Prefab, FindFreePosition(), unitRotation).GetComponent<UnitAI>();
                GameObjectManager.AddToPlayerObjectsList(OwnerPlayer, newUnit.transform);

                ////    
                //var message = new TaskDataMessage {
                //    TaskData = new FollowToPoint_Task.FollowToPoint_TaskData(AssemblyPoint ),
                //    NewQueue = true
                //};
                //newUnit.AddTask(message);

                DequeueUnitData();
            }
        }

    }

    void OnDestroy()
    {
        //     ,     
        if (_taskDataList.Count > 0)
        {
            foreach (CreatingUnit_Data taskData in _taskDataList)
                OwnerPlayer.Money += taskData.Price;
        }
    }

    void DequeueUnitData()
    {
        _currentTaskData=null;
        if (_taskDataList.Count > 0)
        {
            _hasCreatingUnit = true;
            _currentTaskData = _taskDataList[0];
            _timeBeforeSpawn = _currentTaskData.CreatingDuration;
            _taskDataList.RemoveAt(0);
        }
        else
        {
            _hasCreatingUnit = false;
            enabled = false;
        }
    }
    
    Vector3 FindFreePosition()
    {
        Vector3 pos = transform.TransformPoint(_spawnPoint);
        Node defaultSpawnNode = AstarPath.active.GetNearest(pos);

        if (NodeIsWalkable(defaultSpawnNode) && PositionIsFree(pos, 0.5f))
            return pos;


        int width = AstarPath.active.astarData.gridGraph.width;
        Node[] nodes = AstarPath.active.astarData.gridGraph.nodes;
        int centerNodeIndex = defaultSpawnNode.GetNodeIndex();
        int index; 
        int indexX;//        0..GridWidth
        int columnHeight = 0, rowWidth = 0;

        //(   ) 
        for (int i = 0; i < 100; i++)
        {
            columnHeight++;

            indexX = centerNodeIndex % width + i + 1;
            if (indexX < width)
            {
                for (int j = -columnHeight; j <= columnHeight; j++)
                {
                    index = centerNodeIndex - j * width + i + 1;
                    Vector3 nodePos;
                    if (NodePositionIsFree(nodes, index, out nodePos))
                        return nodePos;
                }
            }


            for (int j = -rowWidth; j <= rowWidth; j++)
            {
                indexX = centerNodeIndex % width - j;
                if (indexX >= 0 && indexX < width)
                {
                    index = centerNodeIndex - (i + 1) * width - j;
                    Vector3 nodePos;
                    if (NodePositionIsFree(nodes, index, out nodePos))
                        return nodePos;
                }
            }

            indexX = centerNodeIndex % width - i - 1;
            if (indexX >= 0)
            {
                for (int j = -columnHeight; j <= columnHeight; j++)
                {
                    index = centerNodeIndex + j * width - i - 1;
                    Vector3 nodePos;
                    if (NodePositionIsFree(nodes, index, out nodePos))
                        return nodePos;
                }
            }

            for (int j = -rowWidth; j <= rowWidth; j++)
            {
                indexX = centerNodeIndex % width + j;
                if (indexX >= 0 && indexX < width)
                {
                    index = centerNodeIndex + (i + 1) * width + j;
                    Vector3 nodePos;
                    if (NodePositionIsFree(nodes, index, out nodePos))
                        return nodePos;
                }
            }

            rowWidth++;
        }
        return pos;
    }

    bool NodeIsWalkable(Node node)
    {
        return (node.tags != (int)GridTags.Buildings && node.walkable);
    }

    bool NodePositionIsFree(Node[] nodes, int nodeIndex, out Vector3 nodePos)
    {
        if (nodeIndex >= 0 && nodeIndex < nodes.Length && NodeIsWalkable(nodes[nodeIndex]))
        {
            Vector3 newNodePos = nodes[nodeIndex].position.ToVector3();
            if (PositionIsFree(newNodePos, 0.5f))
            {
                nodePos=newNodePos;
                return true;
            }
        }
        nodePos = Vector3.zero;
        return false;
    }

    bool PositionIsFree(Vector3 point, float radius)
    {
        Collider[] colls = Physics.OverlapSphere(point, radius, GameManager.AllObjectLayers);
        if (colls.Length > 0)
            return false;
        return true;
    }


    //         .  , ,       
    void OnDrawGizmosSelected()
    {
        if (Application.platform == RuntimePlatform.WindowsEditor || Application.platform == RuntimePlatform.OSXEditor)
        {
            Gizmos.color = new Color(0.0f, 0.5f, 1.0f, 0.5f);
            var tr = GetComponent<Transform>();//GetComponent  transform -      
            float radius = 0.3f;

            Vector3 pos = tr.TransformPoint(_spawnPoint);
            pos.y += radius;

            Gizmos.DrawSphere(pos, radius);

            Quaternion rot = new Quaternion();
            rot.eulerAngles = _spawnRotation;
            Vector3 dir;
            rot = tr.rotation * rot;//  

            //forward
            dir = (rot * Vector3.forward).normalized;
            Gizmos.color = new Color(0, 0, 1);
            Gizmos.DrawRay(pos, dir);

            //up
            dir = (rot * Vector3.up).normalized;
            Gizmos.color = new Color(0, 1, 0);
            Gizmos.DrawRay(pos, dir);

            //right
            dir = (rot * Vector3.right).normalized;
            Gizmos.color = new Color(1, 0, 0);
            Gizmos.DrawRay(pos, dir);
        }
    }
}
